//
// sueda
// November, 2014
//

#include "Camera.h"
#include "MatrixStack.h"
#include <iostream>
#define GLM_FORCE_RADIANS
#include "glm/gtx/string_cast.hpp"
#include "glm/gtx/transform.hpp"

using namespace std;

#define MAX_V 20.0f
#define ACCEL -9.8f

static float v0 = 0.0f;
static float mass = 0.028f;

Camera::Camera() :
	aspect(1.0f),
	fovy(30.0f/180.0f*M_PI),
	znear(0.1f),
	zfar(1000.0f),
	position(0.0f, 0.0f, 0.0f)
{
}

Camera::~Camera()
{
	
}

void Camera::mouseClicked(int x, int y)
{
	mousePrev.x = x;
	mousePrev.y = y;
	state = Camera::ROTATE;
}

glm::vec2 Camera::mouseMoved(int x, int y)
{
	glm::vec2 mouseCurr(x, y);
	glm::vec2 dv = mouseCurr - mousePrev;
	switch(state) {
		case Camera::ROTATE:
			rotations += rfactor * dv;
			break;
	}
	mousePrev.x = x;
	mousePrev.y = y;
	return dv;
}

void Camera::setWindowSize(float w, float h)
{
	aspect = w/h;
}

void Camera::update(const bool *keys, const glm::vec2 &mouse, int dt)
{
	float a, v, x; // acceleration, velocity, position

	// How much has the mouse moved?
	//float dx = mouse.x - mousePrev.x;
	//float dy = mouse.y - mousePrev.y;

	// Update camera position and orientation here
	if(keys['w']) {
		//position.z -= 0.5f;
		position.x += sin(rotations.y) * 0.8f;
		position.z -= cos(rotations.y) * 0.8f;
	} 
	if(keys['s']) {
		//position.z += 0.5f;
		position.x -= sin(rotations.y) * 0.5f;
		position.z += cos(rotations.y) * 0.5f;
	}
	if(keys['a']) {
		rotations.y -= 0.017f;
	} 
	if(keys['d']) {
		rotations.y += 0.017f;
	}
	if(keys[' ']) {
		v0 += 0.12f;
	} 
	if(keys['/'] && position.y > 0.0) {
		//position.y -= 0.1f;
	}

	// vehicle y position physics
	a = ACCEL * mass;
	if(v0 > 3.5f)
		v0 -= 0.12f;
	else if(v0 < -4.0f)
		v0 += 0.12f;
	v = v0 + (a * (float)dt * .01);
	if(position.y <= 0.0f && v < 0.0f)
		v = 0.0f;

	x = position.y + (v * (float)dt * .01);
	
	//printf("a = %f, v = %f, x = %f\n", a, v, x);
	
	position.y = x;
	v0 = v;

	// Save last mouse
	//mousePrev = mouse;
}

void Camera::applyProjectionMatrix(MatrixStack *P) const
{
	P->perspective(fovy, aspect, znear, zfar);
}

void Camera::Reset() {
	position = glm::vec3(0.0f, 0.0f, 0.0f);
	rotations = glm::vec2(0.0f, 0.0f);
	v0 = 0.0f;
}

void Camera::applyViewMatrix(MatrixStack *MV) const
{
	// Create the translation and rotation matrices
	glm::mat4 T = glm::translate(position);

	// The matrix C is the product of these matrices
	glm::mat4 C = T; // Also apply rotations here
	// The view matrix is the inverse
	glm::mat4 V = glm::inverse(C);
	// Add to the matrix stack
	MV->rotate(rotations.y, glm::vec3(0.0f, 1.0f, 0.0f));
	MV->multMatrix(V);
	//MV->rotate(rotations.x, glm::vec3(0.0f, 1.0f, 0.0f));
}
