#ifdef __APPLE__
#include <GLUT/glut.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif
#include <cstdlib>
#include <iostream>
#include <vector>
#define GLM_FORCE_RADIANS
#include "glm/gtx/transform.hpp"
#include "GLSL.h"
#include "Camera.h"
#include "Shape.h"
#include "MatrixStack.h"
#include "Image.h"
#include "odeBoxBox.h"

#define NUM_OBJS 180
#define NUM_SHAPES 4

using namespace std;

int time0;
Shape shape[NUM_SHAPES];
glm::vec2 mouse;
bool keyDown[256] = {false};
Camera camera;
bool cull = false;
bool line = false;
int curShape = 0;
float upDownView = 0.0f;
float leftRightView = 0.0f;
float window_W = 400.0f;
float window_H = 400.0f;
static float yaw = 0.0f;
static float pitch = 0.0f;
static bool collision = false;
static glm::vec3 bunnyColor;
static glm::vec3 buildingC;

glm::mat3 T(1.0);

// Vertex buffer data. This data is populated on the CPU and sent to the GPU.
vector<float> positionBuf;
vector<unsigned int> indexBuf;
// Vertex buffer object IDs. These are used by OpenGL to reference the buffers.
GLuint positionBufID;
GLuint indexBufID;
// Model matrix for the plane
glm::mat4 M;

// GLSL program
GLuint pid;
// GLSL handles to various parameters in the shaders
GLint h_P;
GLint h_MV;
GLint h_vertPos;
GLint h_vertNor;
GLint h_vertTexCoords;
GLint h_lightPos;
GLint h_ka;
GLint h_kd;
GLint h_ks;
GLint h_s;

float randFloat(float l, float h)
{
	float r = rand() / (float)RAND_MAX;
	return (1.0f - r) * l + r * h;
}

static GLfloat loc[NUM_OBJS * 3];
static GLfloat color[NUM_OBJS * 3];
static GLfloat rot[NUM_OBJS * 3];
static GLfloat scale_tree[NUM_OBJS];
static GLfloat tree_loc[NUM_OBJS * 3];
static GLfloat cloud_loc[NUM_OBJS * 3];
static GLfloat treeColor[NUM_OBJS * 3];

void printMat(const glm::mat4 &M, const char *name = 0)
{
	if(name) {
		printf("%s=[\n", name);
	}
	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			printf("%- 5.2f ", M[j][i]);
		}
		printf("\n");
	}
	if(name) {
		printf("];");
	}
	printf("\n");
}


void Init_Rand_Arrays() {

	for(int i = 0; i < NUM_OBJS; i++) {
		loc[i * 3 + 0] = randFloat(-1.5f, 1.5f);
		loc[i * 3 + 1] = randFloat(-1.5f, 1.5f);
		loc[i * 3 + 2] = randFloat(-2.0f, 2.0f);

		color[i * 3 + 0] = 0.2f;
		color[i * 3 + 1] = 0.2f;
		color[i * 3 + 2] = 0.2f;

		treeColor[i * 3 + 0] = 0.1f;
		treeColor[i * 3 + 1] = 0.3f;
		treeColor[i * 3 + 2] = 0.1f;

		rot[i * 3 + 0] = randFloat(0.0f, 1.0f);
		rot[i * 3 + 1] = randFloat(0.0f, 1.0f);
		rot[i * 3 + 2] = randFloat(0.0f, 1.0f);

		scale_tree[i] = randFloat(10.0f, 25.0f);

		tree_loc[i * 2 + 0] = randFloat(-200.0f, 200.0f);
		tree_loc[i * 2 + 1] = randFloat(-200.0f, 200.0f);

		cloud_loc[i * 3 + 0] = randFloat(-350.0f, 350.0f);
		cloud_loc[i * 3 + 1] = randFloat(30.0f, 100.0f);
		cloud_loc[i * 3 + 2] = randFloat(-350.0f, 350.0f);
	}
}


void loadScene()
{
	// time
	time0 = 0;
	
	// Populate the world here
	positionBuf.push_back(-1.0f);
	positionBuf.push_back(-1.0f);
	positionBuf.push_back(0.0f);
	positionBuf.push_back(1.0f);
	positionBuf.push_back(-1.0f);
	positionBuf.push_back(0.0f);
	positionBuf.push_back(-1.0f);
	positionBuf.push_back(1.0f);
	positionBuf.push_back(0.0f);
	positionBuf.push_back(1.0f);
	positionBuf.push_back(1.0f);
	positionBuf.push_back(0.0f);
	indexBuf.push_back(0);
	indexBuf.push_back(1);
	indexBuf.push_back(2);
	indexBuf.push_back(3);
	indexBuf.push_back(2);
	indexBuf.push_back(1);
	
	M = glm::translate(glm::vec3(0.0f, 0.0f, -10.0f));
	bunnyColor = glm::vec3(randFloat(0.0f, 1.0f), 
	 randFloat(0.0f, 1.0f), randFloat(0.0f, 1.0f));
	buildingC = glm::vec3(0.2f, 0.2f, 0.2f);
	
	shape[0].load("meshes/bunny.obj");
	shape[1].load("meshes/cube.obj");
	shape[2].load("meshes/sphere.obj");
	shape[3].load("meshes/tree.obj");
	shape[4].load("meshes/building1.obj");
}

void initGL()
{
	//////////////////////////////////////////////////////
	// Initialize GL for the whole scene
	//////////////////////////////////////////////////////
	
	// Set background color
	glClearColor(0.0f, 0.55f, 0.8f, 1.0f);
	// Enable z-buffer test
	glEnable(GL_DEPTH_TEST);
	
	//////////////////////////////////////////////////////
	// Intialize the shaders
	//////////////////////////////////////////////////////
	shape[0].init();
	shape[1].init();
	shape[2].init();
	shape[3].init();
	shape[4].init();
	
	// Create shader handles
	string vShaderName = "simple_vert.glsl";
	string fShaderName = "simple_frag.glsl";
	GLint rc;
	GLuint VS = glCreateShader(GL_VERTEX_SHADER);
	GLuint FS = glCreateShader(GL_FRAGMENT_SHADER);
	
	// Read shader sources
	const char *vshader = GLSL::textFileRead(vShaderName.c_str());
	const char *fshader = GLSL::textFileRead(fShaderName.c_str());
	glShaderSource(VS, 1, &vshader, NULL);
	glShaderSource(FS, 1, &fshader, NULL);
	
	// Compile vertex shader
	glCompileShader(VS);
	GLSL::printError();
	glGetShaderiv(VS, GL_COMPILE_STATUS, &rc);
	GLSL::printShaderInfoLog(VS);
	if(!rc) {
		printf("Error compiling vertex shader %s\n", vShaderName.c_str());
	}
	
	// Compile fragment shader
	glCompileShader(FS);
	GLSL::printError();
	glGetShaderiv(FS, GL_COMPILE_STATUS, &rc);
	GLSL::printShaderInfoLog(FS);
	if(!rc) {
		printf("Error compiling fragment shader %s\n", fShaderName.c_str());
	}
	
	// Create the program and link
	pid = glCreateProgram();
	glAttachShader(pid, VS);
	glAttachShader(pid, FS);
	glLinkProgram(pid);
	GLSL::printError();
	glGetProgramiv(pid, GL_LINK_STATUS, &rc);
	GLSL::printProgramInfoLog(pid);
	if(!rc) {
		printf("Error linking shaders %s and %s\n", vShaderName.c_str(), fShaderName.c_str());
	}
	
	h_vertPos = GLSL::getAttribLocation(pid, "vertPos");
	h_vertNor = GLSL::getAttribLocation(pid, "vertNor");
	h_vertTexCoords = GLSL::getAttribLocation(pid, "vertTexCoords");
	h_P = GLSL::getUniformLocation(pid, "P");
	h_MV = GLSL::getUniformLocation(pid, "MV");
	h_lightPos = GLSL::getUniformLocation(pid, "lightPos");
	h_ka = GLSL::getUniformLocation(pid, "ka");
	h_kd = GLSL::getUniformLocation(pid, "kd");
	h_ks = GLSL::getUniformLocation(pid, "ks");
	h_s = GLSL::getUniformLocation(pid, "s");
	//h_texture0 = GLSL::getUniformLocation(pid, "texture0");


	//////////////////////////////////////////////////////
	// Intialize textures

	
	//////////////////////////////////////////////////////
	// Initialize the vertex buffers
	//////////////////////////////////////////////////////
	
	// Send the position array to the GPU
	glGenBuffers(1, &positionBufID);
	glBindBuffer(GL_ARRAY_BUFFER, positionBufID);
	glBufferData(GL_ARRAY_BUFFER, positionBuf.size()*sizeof(float), &positionBuf[0], GL_STATIC_DRAW);
	
	// Send the index array to the GPU
	glGenBuffers(1, &indexBufID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBufID);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indexBuf.size()*sizeof(unsigned int), &indexBuf[0], GL_STATIC_DRAW);
	
	// Unbind the arrays
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	// Check GLSL
	GLSL::checkVersion();
}

void reshapeGL(int w, int h)
{
	// Set view size
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	// Set camera aspect ratio
	camera.setWindowSize(w, h);
	window_W = (float)w;
	window_H = (float)h;
}

void drawGL()
{
	// Clear buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	// Enable backface culling
	if(cull) {
		glEnable(GL_CULL_FACE);
	} else {
		glDisable(GL_CULL_FACE);
	}
	if(line) {
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	} else {
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	
	// Create matrix stacks
	MatrixStack P, MV, R;

	glm::mat4 bunnyM, cloudM, treeM, buildingM;
	glm::vec3 sides1(10.0f, 10.0f, 10.0f);
	glm::vec3 sides2;

	R.loadIdentity();

	// Apply camera transforms
	P.pushMatrix();
	camera.applyProjectionMatrix(&P);
	
	// Bind the program
	glUseProgram(pid);
	glUniform3fv(h_lightPos, 1, glm::value_ptr(glm::vec3(2.0f, 2.0f, 2.0f)));
	glUniform3fv(h_kd,  1, glm::value_ptr(glm::vec3(0.4f, 0.3f, 0.3f)));
	glUniform3fv(h_ks, 1, glm::value_ptr(glm::vec3(0.5f, 0.4f, 0.4f)));
	glUniform1f(h_s, 2000.0f);

	// Send projection matrix to the GPU
	glUniformMatrix4fv(h_P, 1, GL_FALSE, glm::value_ptr( P.topMatrix()));
	
	// Enable and bind position array
	GLSL::enableVertexAttribArray(h_vertPos);
	glBindBuffer(GL_ARRAY_BUFFER, positionBufID);
	glVertexAttribPointer(h_vertPos, 3, GL_FLOAT, GL_FALSE, 0, 0);

	// Bind index array
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBufID);

	MV.pushMatrix();

	// bunny code
	MV.pushMatrix();
	MV.translate(glm::vec3(0.0f, -0.2f, -2.0f));
	MV.scale(glm::vec3(0.25f, 0.25f, 0.25f));
	if(keyDown['a']) {
		MV.rotate(0.6, glm::vec3(0.0f, 0.0f, 1.0f));
	}
	if(keyDown['d']) {
		MV.rotate(-0.6, glm::vec3(0.0f, 0.0f, 1.0f));
	}
	if(keyDown['w']) {
		MV.rotate(-0.6, glm::vec3(1.0f, 0.0f, 0.0f));
	}
	if(keyDown['s']) {
		MV.rotate(0.6, glm::vec3(1.0f, 0.0f, 0.0f));
	}
	MV.rotate(-M_PI / 2.0, glm::vec3(0.0f, 1.0f, 0.0f));
	glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
	glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(bunnyColor.x, 
	 bunnyColor.y, bunnyColor.z)));
	shape[0].draw(h_vertPos, h_vertNor, h_vertTexCoords);
	bunnyM = MV.topMatrix();
	MV.popMatrix();

	camera.applyViewMatrix(&MV);

	// Apply model matrix and send to GPU
	MV.pushMatrix();
	MV.translate(glm::vec3(0.0f, -1.0f, 0.0f));
	MV.scale(glm::vec3(600.0f, 0.01f, 600.0f));
	glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
	glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(0.2f, 0.5f, 0.2f)));
	shape[1].draw(h_vertPos, h_vertNor, h_vertTexCoords);
	MV.popMatrix();


	sides2 = glm::vec3(8.0f, 8.0f, 130.0f);
	sides1 = glm::vec3(8.0f, 8.0f, 130.0f);
	// draw building	
	MV.pushMatrix();
	MV.translate(glm::vec3(60.0f, -2.0f, -150.0f));
	MV.scale(glm::vec3(.4f, .4f, .4f));
	MV.rotate(-M_PI / 2.0, glm::vec3(1.0f, 0.0f, 0.0f));
	glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
	buildingM = MV.topMatrix();
	Contacts contacts = odeBoxBox(bunnyM, sides1, buildingM, sides2);
	if(contacts.positions.size() > 0) {
		buildingC = glm::vec3(randFloat(0.0f, 1.0f), 
	 randFloat(0.0f, 1.0f), randFloat(0.0f, 1.0f));
	}
	glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(buildingC.x, buildingC.y,
	 buildingC.z)));
	shape[4].draw(h_vertPos, h_vertNor, h_vertTexCoords);
	MV.popMatrix();
		
	sides2 = glm::vec3(30.0f, 30.0f, 30.0f);
	sides1 = glm::vec3(30.0f, 30.0f, 30.0f);
	for(int i = 0; i < 15; i++) {

		MV.pushMatrix();
		MV.translate(glm::vec3(tree_loc[i * 2], 
		 0.32f * scale_tree[i], tree_loc[i * 2 + 1]));
		MV.rotate(-M_PI / 2.0, glm::vec3(1.0f, 0.0f, 0.0f));
		MV.scale(glm::vec3(scale_tree[i], scale_tree[i], scale_tree[i]));
		glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
		glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(treeColor[3 * i + 0], 
		 treeColor[3 * i + 1], treeColor[3 * i + 2])));
		shape[3].draw(h_vertPos, h_vertNor, h_vertTexCoords);
		treeM = MV.topMatrix();
		MV.popMatrix();

		Contacts contacts = odeBoxBox(bunnyM, sides1, treeM, sides2);
		if(contacts.positions.size() > 0) {
			collision = true;
			treeColor[3 * i + 0] = randFloat(0.0f, 1.0f);
			treeColor[3 * i + 1] = randFloat(0.0f, 1.0f);
			treeColor[3 * i + 2] = randFloat(0.0f, 1.0f);
		}
	}
	

	// draw square clouds
	sides2 = glm::vec3(30.0f, 30.0f, 30.0f);
	sides1 = glm::vec3(30.0f, 30.0f, 30.0f);
	for(int i = 0; i < NUM_OBJS / 2; i++) {

		MV.pushMatrix();
		MV.translate(glm::vec3(cloud_loc[i * 3], cloud_loc[i * 3 + 1], 
		 cloud_loc[i * 3 + 2]));
		MV.rotate(-M_PI / 2.0, glm::vec3(1.0f, 0.0f, 0.0f));
		MV.scale(glm::vec3(10.0f, 10.0f, 10.0f));
		MV.rotate(M_PI, glm::vec3(rot[0 + 3 * i], rot[1 + 3 * i], 
	    rot[2 + 3 * i]));
		glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
		glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(color[3 * i],
		 color[3 * i + 1], color[3 * i + 2])));
		shape[1].draw(h_vertPos, h_vertNor, h_vertTexCoords);
		cloudM = MV.topMatrix();
		MV.popMatrix();

		Contacts contacts = odeBoxBox(bunnyM, sides1, cloudM, sides2);
		if(contacts.positions.size() > 0) {
			collision = true;
			color[i * 3 + 0] = randFloat(0.0f, 1.0f);
			color[i * 3 + 1] = randFloat(0.0f, 1.0f);
			color[i * 3 + 2] = randFloat(0.0f, 1.0f);
		}
	}

	// draw sphere clouds
	sides2 = glm::vec3(25.0f, 25.0f, 25.0f);
	sides1 = glm::vec3(25.0f, 25.0f, 25.0f);
	for(int i = NUM_OBJS / 2; i < NUM_OBJS; i++) {

		MV.pushMatrix();
		MV.translate(glm::vec3(cloud_loc[i * 3], cloud_loc[i * 3 + 1], 
		 cloud_loc[i * 3 + 2]));
		MV.rotate(-M_PI / 2.0, glm::vec3(1.0f, 0.0f, 0.0f));
		MV.scale(glm::vec3(10.0f, 10.0f, 10.0f));
		MV.rotate(M_PI, glm::vec3(rot[0 + 3 * i], rot[1 + 3 * i], 
	    rot[2 + 3 * i]));
		glUniformMatrix4fv(h_MV, 1, GL_FALSE, glm::value_ptr(MV.topMatrix()));
		glUniform3fv(h_ka,  1, glm::value_ptr(glm::vec3(color[3 * i],
		 color[3 * i + 1], color[3 * i + 2])));
		shape[2].draw(h_vertPos, h_vertNor, h_vertTexCoords);
		cloudM = MV.topMatrix();
		MV.popMatrix();

		Contacts contacts = odeBoxBox(bunnyM, sides1, cloudM, sides2);
		if(contacts.positions.size() > 0) {
			collision = true;
			color[i * 3 + 0] = randFloat(0.0f, 1.0f);
			color[i * 3 + 1] = randFloat(0.0f, 1.0f);
			color[i * 3 + 2] = randFloat(0.0f, 1.0f);
		}
	}

	// Disable and unbind
	GLSL::disableVertexAttribArray(h_vertPos);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	// Unbind textures
	glActiveTexture(GL_TEXTURE0 + 2);
	glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTexture(GL_TEXTURE0 + 1);
	glBindTexture(GL_TEXTURE_2D, 0);
	glActiveTexture(GL_TEXTURE0 + 0);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	// Unbind the program
	glUseProgram(0);

	// Pop stacks
	MV.popMatrix();
	P.popMatrix();

	// Double buffer
	glutSwapBuffers();
}

void passiveMotionGL(int x, int y)
{
	mouse.x = x;
	mouse.y = y;
}

void mouseGL(int button, int state, int x, int y)
{
	int modifier = glutGetModifiers();
	camera.mouseClicked(x, y);
}

void motionGL(int x, int y)
{

	glm::vec2 change = camera.mouseMoved(x, y);
	
	// obtain pitch and yaw rotations
	pitch += change.x;
	yaw += change.y;

	// Refresh screen
	glutPostRedisplay();
}

void keyboardGL(unsigned char key, int x, int y)
{
	keyDown[key] = true;
	switch(key) {
		case 27:
			// ESCAPE
			exit(0);
			break;
		case 'w':
			upDownView += 2.0f;
			//camera.update(keyDown, mouse);
			break;
		case 's':
			upDownView -= 2.0f;
			//camera.update(keyDown, mouse);
			break;
		case 'a':
			pitch -= 0.05f;
			//camera.update(keyDown, mouse);
			break;
		case 'd':
			pitch += 0.05f;
			//camera.update(keyDown, mouse);
			break;
		case '/':
			//camera.update(keyDown, mouse);
			break;
		case ' ':
			//camera.update(keyDown, mouse);
			break;
		case 'r':
			camera.Reset();
			collision = false;
		break;
	}
}

void keyboardUpGL(unsigned char key, int x, int y)
{
	keyDown[key] = false;
}

void idleGL()
{
	curShape = 0;
	int time1 = glutGet(GLUT_ELAPSED_TIME);
	int dt = time1 - time0;
	// Update every 60Hz

	if(collision) {
		bunnyColor = glm::vec3(randFloat(0.0f, 1.0f), 
	    randFloat(0.0f, 1.0f), randFloat(0.0f, 1.0f));
		collision = false;
	}
	if(dt > 1000.0/60.0) {
		time0 = time1;
		camera.update(keyDown, mouse, dt);
		glutPostRedisplay();
	}
}

int main(int argc, char **argv)
{
	Init_Rand_Arrays();
	glutInit(&argc, argv);
	glutInitWindowSize(400, 400);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_MULTISAMPLE);
	glutCreateWindow("Matt Mitchell");
	glutMouseFunc(mouseGL);
	glutMotionFunc(motionGL);
	glutPassiveMotionFunc(passiveMotionGL);
	glutKeyboardFunc(keyboardGL);
	glutKeyboardUpFunc(keyboardUpGL);
	glutReshapeFunc(reshapeGL);
	glutDisplayFunc(drawGL);
	glutIdleFunc(idleGL);
	loadScene();
	initGL();
	glutMainLoop();
	return 0;
}
